import os
from os import path
import matplotlib.pyplot as plt
#import ipywidgets as widgets

from IPython.display import display, HTML
#from IPython.core.magic import register_cell_magic

#import pickle
import panel as pn

import pandas as pd
import seaborn as sns

#todecomment for final version
import warnings
warnings.simplefilter("ignore", UserWarning)

from frangiPANe.snakemake_scripts import frangiPANe

def stat(df, col_stat):
    df_stat = df.agg({col_stat: ['mean', 'min', 'max']})
    return df_stat[col_stat]['min'], df_stat[col_stat]['max'], round(df_stat[col_stat]['mean'], 2)

def group_stat(df, col_group, col_stat):
    return df.groupby(col_group).agg({col_stat: ['mean', 'min', 'max']})


def format_60(txt):
    output = ''
    cpt = 0
    for c in txt:
        if cpt < 60:
            output += c
            cpt += 1
        else:
            output += '\n' + c
            cpt = 1
    return (output)

def format_stat(min, max, mean):
    return f"""<br />\n
* min - max : {min} - {max} %
* mean : {mean}"""


def format_stat_int(min, max, mean):
    return f"""<br />\n
* min - max : {int(min)} - {int(max)}
* mean : {int(mean)}"""

def convert_mb(size):
    return round(size / 1000000, 2)

def get_genome_size(reference_genome):

    from Bio import SeqIO

    #  list to collect for each chromosome, chromosome size
    genome = []

    # read fasta genome reference
    FastaFile = open(reference_genome, 'r')
    for rec in SeqIO.parse(FastaFile, 'fasta'):
        seq_name = rec.id
        seq_len = len(rec.seq)
        # add chromosome name and size into genome list
        genome.append([seq_name, seq_len])
    FastaFile.close()

    # Create pandaframe from list
    df_genome = pd.DataFrame(genome, columns=["chr", "size", ])

    ### calculate genome size
    return df_genome['size'].sum()


def dashboard_genome(reference_genome, output_png, output_csv):

    from Bio import SeqIO

    #  list to collect for each chromosome, chromosome size
    genome = []

    # read fasta genome reference
    FastaFile = open(reference_genome, 'r')
    for rec in SeqIO.parse(FastaFile, 'fasta'):
        seq_name = rec.id
        seq_len = len(rec.seq)
        # add chromosome name and size into genome list
        genome.append([seq_name, seq_len])
    FastaFile.close()

    # Create pandaframe from list
    df_genome = pd.DataFrame(genome, columns=["chr", "size", ])

    ### Generate ideogram file used to generate map of contigs anchored on genome
    df_ideogram = df_genome
    df_ideogram['start'] = 0
    df_ideogram['end'] = df_ideogram['size'] - 1
    df_ideogram['name'] = df_ideogram['chr']
    df_ideogram['gieStain'] = 'gvar'

    ### calculate genome size
    total_genome = df_genome['size'].sum()
    df_genome['size'] = round(df_genome['size'].div(1000000), 2)

    ### plot genome size of each chromosome
    # Create a visualization
    plt.figure(figsize=(8, 6))
    ax = sns.barplot(
        data=df_genome,
        x="chr", y="size"
    )
    ax.set(xlabel="Chr", ylabel="Size (Mb)")
    plt.savefig(output_png)
    plt.close()
 #   stat = f"## Genome size (Mb) : {df_genome['size'].sum()}"

    # drop column size of df_ideogram and write the df into the ideogram file
    df_ideogram.drop('size', 1, inplace=True)
    df_ideogram.to_csv(output_csv, sep='\t', index=False, header=False)

    return total_genome

def dashboard_genome2(genome_size, output_png, output_csv):

    col_list= ["chr", "start", "size","id","type"]
    # Create pandaframe from list
    #df_genome = pd.DataFrame(genome, columns=["chr", "size", ])
    df_genome = pd.read_csv(output_csv, index_col=False, header=None, sep="\t", names=col_list)

    stat = f"## Genome size (Mb) : {genome_size}"
    dashboard_title = '# Reference genome file '

    pn.extension()

    dashboard = pn.Column(dashboard_title, stat, pn.Row(output_png, pn.panel(df_genome, styles={'background': 'WhiteSmoke'})),
                          sizing_mode='stretch_both').servable()
    display(dashboard)

def dashboard_group(df):
    pn.extension()

    # create a class containing an species selector drop-down, various plots, and a data table output
    class GroupDashboard(param.Parameterized):
        # drop down selector widget containing the list of species
        # Animal = param.ObjectSelector(default='Goat', objects=list(df.Animal.unique()))
        Species = param.ObjectSelector(default=df.Species.unique().any(), objects=list(df.Species.unique()))

        # create data set containing only the data applicable to the species  in the drop down selector
        def get_data(self):
            class_df = df[(df.Species == self.Species)].copy()
            return class_df

        # seaborn box plot for the chosen animal
        def box_view(self):
            data = df
            plt.figure()  # figsize=(4, 3))
            ax = sns.countplot(y=data['Species'], palette="hls")
            plt.close()
            return ax.figure

        # table of data for the chosen animal
        def table_view(self):
            data = self.get_data()
            return data

    species_list = df.Species.unique().tolist()
    stat = f"## Samples : {df['sample'].count()} \n#### Group : {' '.join(species_list)}\n "

    # create an instance of the class
    rd = GroupDashboard(name='')

    # create a title for the dashboard
    dashboard_title = '# Population Group'

    # create some text describing the dashboard
    dashboard_desc = 'Some statistics and list sample by group '

    # create a dashboard, defining the layout as one column containing the
    # dashboard title, dashboard description, drop down selector,
    # box plot, and data table
    dashboard = pn.Column(dashboard_title,
                          # dashboard_desc,
                          pn.Row(stat, rd.box_view),  # box plot
                          rd.param,
                          pn.Row(rd.table_view, scroll=True, width=800, height=400),  # data table
                          sizing_mode='stretch_both', background='WhiteSmoke', scroll=True
                          )

    # show the dashboard with the data embedded ,
    # (for using in an html download of the notebook so that
    # no 'callback' is required from the browser to get the data)
    dashboard.embed()


def dashboard_fastq(csv, total_size, group_file, output_dir):

    id_dict,df_group = frangiPANe.read_group_file(group_file)
    col_list = ["sample", "file", "reads", "len", "len_mean", "len_min", "qual_min", "qual_max", "qual_mean", "%A",
                "%C", "%G", "%T", "%N", "total_bases"]
    df_fastq_stat = pd.read_csv(csv, index_col=False, sep="\t", usecols=col_list)
    df_fastq_stat.sort_values(by=['sample', 'file'], inplace=True)

    df_stat_merged = pd.merge(left=df_group, right=df_fastq_stat, left_on='sample', right_on='sample')

    species_count = df_stat_merged.groupby('sample')['Species'].unique().value_counts()

    table_sp = f"""| Samples by group  |  |\n|:---|:---:|\n"""
    for key, item in species_count.items():
        table_sp += f"| {key[0]} : | {int(item)} samples |\n"

    sns.set_style("darkgrid")

    print(f"--------{output_dir}/sample_total_pb.png")
    # Plot total bases pb by sample
    df_merge = df_stat_merged.groupby('sample')['total_bases'].sum()
    plt.figure()  # figsize=(10, 5))
    total = df_merge.plot(kind='bar', color='#66C2A5')
    total.set(xlabel="sample", ylabel="Total (Mb)")
    total.set_title("Total pb sequenced by sample")
    plt.savefig(os.path.join(output_dir,f"sample_total_pb.png"))
    plt.close()

    total_desc = f"""### About population
* {len(df_stat_merged['sample'].unique())} samples
* Total base sequenced : {int(df_stat_merged['total_bases'].sum() / 1000000)} Mb"""

    df_grouped = df_stat_merged.groupby('Species')['total_bases'].sum().div(1000000)

    table = f"""| Pb sequenced by groups |  |\n|:---|:---:|\n"""
    for key, item in df_grouped.items():
        table += f"| {key} : | {int(item)} Mb |\n"

    # Coverage
    df_merge = df_stat_merged.groupby('sample')['total_bases'].sum().div(total_size)
    plt.figure()  # figsize=(10, 5))
    x = df_merge.plot(kind='bar', color='#66C2A5')
    x.set(xlabel="sample", ylabel="X")
    x.set_title("Sequencing coverage by sample")
    plt.savefig(os.path.join(output_dir,f"sample_sequencingcoverage.png"))
    plt.close()

    # Plot length read
    # df_read = df_stat_merged[["file","len_min","len","len_mean"]]
    plt.figure()  # figsize=(10, 5))
    read = sns.scatterplot(x='file', y='len_mean', data=df_stat_merged, linewidth=2, palette="colorblind")
    read.set(xlabel="file", ylabel="read length (pb)")
    read.set_title("Read length over samples")
    plt.savefig(os.path.join(output_dir,f"sample_reads_length.png"))
    plt.close()


    # Plot quality
    # df_qual = df_fstat_merged[["file","qual_min","qual_max","qual_mean"]]
    plt.figure()  # figsize=(10, 5))
    qual = sns.scatterplot(x='file', y='qual_mean', data=df_stat_merged, linewidth=2.5, palette="bright")
    qual.set(xlabel="file", ylabel="qual mean")
    qual.set_title("Quality mean over samples")
    plt.savefig(os.path.join(output_dir,f"sample_reads_qual.png"))
    plt.close()


    # Plot base ratio
    df_ratio = df_stat_merged[["file", "%A", "%C", "%G", "%T", "%N"]]
    plt.figure()  # figsize=(10, 5))
    base = sns.lineplot(x='file', y='value', hue='variable', data=pd.melt(df_ratio, 'file'),
                        linewidth=2, markersize=8, palette="hls", marker='o', linestyle='dashed')
    base.set(xlabel="file", ylabel="base ratio")
    base.set_title("Base ratio (A,T,C,G,N)")
    plt.legend('ACGTN', ncol=1, loc='upper left')
    plt.savefig(os.path.join(output_dir,f"sample_reads_base_ratio.png"))
    plt.close()

def dashboard_flagstat(stat_file, group_file, output_png):

    id_dict,df_group = frangiPANe.read_group_file(group_file)
    
    df_bam_stat = pd.read_csv(stat_file, index_col=False, sep=",")
    df_bam_stat.sort_values(by=['sample'], inplace=True)

    df_stat_merged = pd.merge(left=df_bam_stat, right=df_group, left_on='sample', right_on='sample')

    # check pop number
    species_count = df_stat_merged.groupby('sample')['Species'].unique().value_counts()
    table_sp = ""#f"""| Samples by group  |  |\n|:---|:---:|\n"""
    for key, item in species_count.items():
        table_sp += f"| {key[0]} : | {int(item)} samples |\n"
    print(table_sp)

    # Stat map
    mapped_min, mapped_max, mapped_mean = stat(df_stat_merged, 'MAPPED')
    mapped_title = "<br /><hr><br />\n### Reads mapped"
    mapped_desc = format_stat(mapped_min, mapped_max, mapped_mean)
    mapped_desc_sp = group_stat(df_stat_merged, 'Species', 'MAPPED')

    paired_min, paired_max, paired_mean = stat(df_stat_merged, 'PAIRED')
    paired_title = "<br /><hr><br />\n### Reads mapped properly paired"
    paired_desc = format_stat(paired_min, paired_max, paired_mean)
    paired_desc_sp = group_stat(df_stat_merged, 'Species', 'PAIRED')

    # stat unmapped
    unmapped_min, unmapped_max, unmapped_mean = stat(df_stat_merged, 'UNMAPPED')
    unmapped_title = "<br /><hr><br />\n### Reads unmapped"
    unmapped_desc = format_stat(unmapped_min, unmapped_max, unmapped_mean)
    unmapped_desc_sp = group_stat(df_stat_merged, 'Species', 'UNMAPPED')

    plt.figure()  # figsize=(12, 5))
    ratio = sns.scatterplot(x='sample', y='value', hue='variable', data=pd.melt(df_bam_stat, 'sample'))
    ratio.set(xlabel="sample", ylabel="ratio")
    ratio.set_title("Read mapped ratio ")
    plt.legend(bbox_to_anchor=(1.01, 1), borderaxespad=0);
    plt.savefig(output_png)
    plt.close()

def dashboard_assembly(stat_file, group_file):

    id_dict,df_group = frangiPANe.read_group_file(group_file)
  
    # read stat file
    df_stat = pd.read_csv(stat_file, index_col=False, sep="\t")

    # extract  sample name from the column "filename"
    df_stat['sample'] = df_stat.filename.apply(lambda x: os.path.basename(x))
    df_stat['sample'] = df_stat['sample'].str.split('_', expand=True)[0]

    # Merge the dataframes "stat" and "group" into a new dataframe
    df_stat_merged = pd.merge(left=df_stat, right=df_group, left_on='sample', right_on='sample')
    df_stat_merged = df_stat_merged[
        ["sample", "number", "mean_length", "shortest", "longest", "total_length", "Species"]]
    df_stat_merged.sort_values(by=['sample'], inplace=True)

    ctg_min, ctg_max, ctg_mean = stat(df_stat_merged, 'number')
    pb_min, pb_max, pb_mean = stat(df_stat_merged, 'total_length')
    ctg_pb_mean = df_stat_merged.mean_length.mean()
    ctg_pb_max = df_stat_merged.longest.max()
    ctg_pb_min = df_stat_merged.shortest.min()
    df_stat_merged = df_stat_merged.rename(
        columns={"number": "Contigs number", "shortest": "Shortest (pb)", "longest": "Longuest (pb)",
                 "total_length": "total_length (pb)"})

    # Stat contigs numbers
    ctg_title = "<br /><hr><br />\n### Contigs number"
    ctg_desc = format_stat_int(ctg_min, ctg_max, ctg_mean)
    ctg_desc_sp = group_stat(df_stat_merged, 'Species', 'Contigs number')

    # Stat contigs length
    ctg_pb_title = f"<br /><hr><br />\n### Contigs size"
    ctg_pb_desc = format_stat(convert_mb(pb_min), convert_mb(pb_max), convert_mb(pb_mean))
    ctg_pb_sp = group_stat(df_stat_merged, 'Species', 'total_length (pb)')

    # Stat contogh length  ctg_pb_mean
    ctg_ln_title = f"<br /><hr><br />\n### Contigs length"
    ctg_ln_desc = format_stat_int(ctg_pb_min, ctg_pb_max, ctg_pb_mean)

    # plt.figure()#figsize=(12, 5))
    # ratio = sns.scatterplot(x='sample', y='value', hue='variable', data=pd.melt(df_stat_merged, 'sample'))
    # ratio.set(xlabel="sample", ylabel="Contigs number")
    # ratio.set_title("Contigs number")
    # plt.legend(bbox_to_anchor=(1.01, 1), borderaxespad=0);
    # plt.close()
